/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/* YIPS @(#)$Id: handler.c,v 1.1.1.1.2.3 1998/07/07 14:04:07 sakane Exp $ */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>

#include <net/pfkeyv2.h>
#include <netkey/keydb.h>

#include <netinet/in.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include "var.h"
#include "vmbuf.h"
#include "misc.h"
#include "schedule.h"
#include "cfparse.h"
#include "isakmp.h"
#include "oakley.h"
#include "ipsec_doi.h"
#include "crypto.h"
#include "pfkey.h"
#include "admin.h"
#include "extern.h"
#include "debug.h"

char *local_secret;
vchar_t oakley_prime768;
vchar_t oakley_prime1024;

int sockIs[256], n_soI = 1;
int portI = PORT_ISAKMP;
int portIr = PORT_ISAKMP;
struct sockaddr *myaddr[16];

struct isakmp_ph1tab ph1tab;

int isakmp_handler(int);
int isakmp_init(void);
u_char *isakmp_pindex(isakmp_index *, msgid_t *);
void isakmp_close(void);
int isakmp_send(struct isakmp_ph1 *, vchar_t *);
int isakmp_resend_ph1(struct sched *);
int isakmp_timeout_ph1(struct sched *);
int isakmp_resend_ph2(struct sched *);
int isakmp_timeout_ph2(struct sched *);
int isakmp_expire(struct sched *);
static int isakmp_do_expire(struct isakmp_ph1 *);

struct isakmp_ph1 *isakmp_new_ph1(isakmp_index *);
int isakmp_free_ph1(struct isakmp_ph1 *);
struct isakmp_ph1 *isakmp_ph1byindex(isakmp_index *);
struct isakmp_ph1 *isakmp_ph1byindex0(isakmp_index *);
struct isakmp_ph1 *isakmp_ph1byaddr(struct sockaddr *);
struct isakmp_ph2 *isakmp_new_ph2(struct isakmp_ph1 *, msgid_t *);
int isakmp_free_ph2(struct isakmp_ph2 *);
struct isakmp_ph2 *isakmp_ph2bymsgid(struct isakmp_ph1 *, msgid_t *);

int isakmp_pfkey(struct pfkey_st *);
int isakmp_new_queue(struct pfkey_st *, struct sockaddr *);
int isakmp_pfkey_check(struct sched *);
int isakmp_pfkey_over(struct sched *);

struct isakmp_conf *isakmp_cfbypeer(struct sockaddr *);
struct isakmp_conf *isakmp_cfbyaddr(struct sockaddr *, struct sockaddr *);
struct isakmp_proxy *isakmp_proxybyaddr(struct sockaddr *, struct sockaddr *);

vchar_t *isakmp_dump_sa(void);

extern int isakmp_main(vchar_t *, struct sockaddr *);
extern struct isakmp_ph1 *isakmp_begin_ident(struct isakmp_conf *);
extern int isakmp_begin_quick(struct isakmp_ph1 *, struct pfkey_st *);

/*
 * isakmp packet handler
 */
int
isakmp_handler(sockI)
	int sockI;
{
	struct isakmp isakmp;
	struct sockaddr from;
	int fromlen = sizeof(from);
	int len;
	vchar_t *buf;
	int error = -1;

	/* read message by MSG_PEEK */
	while ((len = recvfrom(sockI, (char *)&isakmp, sizeof(isakmp),
	                    MSG_PEEK, &from, &fromlen)) < 0) {
		if (errno == EINTR) continue;
		plog("isakmp_handler", "recvfrom (%s)\n", strerror(errno));
		goto end;
	}

	/* check isakmp header length */
	if (len < sizeof(isakmp)) {
		plog2(&from, "isakmp_handler", "received invalid header length.\n");
		if ((len = recvfrom(sockI, (char *)&isakmp, sizeof(isakmp),
		                    0, &from, &fromlen)) < 0) {
			plog("isakmp_handler", "recvfrom (%s)\n", strerror(errno));
		}
		goto end;
	}

	/* read real message */
	if ((buf = vmalloc(ntohl(isakmp.len))) == 0) {
		plog("isakmp_handler", "vmalloc (%s)\n", strerror(errno)); 
		goto end;
	}

	while ((len = recvfrom(sockI, buf->v, buf->l,
	                    0, &from, &fromlen)) < 0) {
		if (errno == EINTR) continue;
		plog("isakmp_handler", "recvfrom (%s)\n", strerror(errno));
		goto end;
	}

	if (len != buf->l) {
		plog2(&from, "isakmp_handler", "received invalid length, why ?\n");
		goto end;
	}

	YIPSDEBUG(DEBUG_NET,
		INET_NTOP(&from, _addr1_);
		plog("isakmp_handler",
			"%d bytes message has been received from %s[%d].\n",
			len, _addr1_, ntohs(_INPORTBYAF(&from))));
	YIPSDEBUG(DEBUG_DNET, pvdump(buf));

	/* check sender whether allow or not */

	/* XXX: I don't know how to check isakmp half connection. */

	/* isakmp main routine */
	if (isakmp_main(buf, &from) != 0) goto end;

	error = 0;

end:
	vfree(buf);

	return(error);
}

/* %%%
 *
 */
int
isakmp_init()
{
	int tmp, i;

	/* set DH MODP */
	oakley_prime768.v = strtob(OAKLEY_PRIME_MODP768, 16, &oakley_prime768.l);
	oakley_prime1024.v = strtob(OAKLEY_PRIME_MODP1024, 16, &oakley_prime1024.l);

	/* initialize a isakmp status table */
	memset((char *)&ph1tab, 0, sizeof(ph1tab));

	/* initialize a port of ISAKMP to send */
    {
	u_int a = htonl(INADDR_ANY);

	/* for V4 */
	if ((myaddr[0]
	    = CALLOC(sizeof(struct sockaddr_in), struct sockaddr *)) == 0) {
		plog("isakmp_init", "calloc (%s)\n", strerror(errno));
		goto err;
	}
	memset(myaddr[0], 0, sizeof(struct sockaddr_in));
	myaddr[0]->sa_len = sizeof(struct sockaddr_in);
	myaddr[0]->sa_family = PF_INET;
	((struct sockaddr_in *)myaddr[0])->sin_port = htons(portI);
	memcpy(&((struct sockaddr_in *)myaddr[0])->sin_addr.s_addr, &a,
		sizeof(a));
    }

	/* initialize a port of ISAKMP */
	for (i = 0; i < n_soI; i++) {
		if ((sockIs[i] = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
			plog("isakmp_init", "socket (%s)\n", strerror(errno));
			goto err;
		}

		if (setsockopt(sockIs[i], SOL_SOCKET, SO_REUSEPORT,
		               (void *)&tmp, sizeof(tmp)) < 0) {
			plog("isakmp_init", "setsockopt (%s)\n", strerror(errno));
			goto err;
		}

		if (bind(sockIs[i], (struct sockaddr *)myaddr[i],
		         sizeof(*myaddr[i])) < 0) {
			plog("isakmp_init", "bind (%s)\n", strerror(errno));
			goto err;
		}

		YIPSDEBUG(DEBUG_INFO,
			INET_NTOP(myaddr[i], _addr1_);
			plog("isakmp_init",
				"using %s[%d] as the port of isakmp.\n",
				_addr1_, ntohs(_INPORTBYAF(myaddr[i]))));
	}

	return(0);

err:
	isakmp_close();
	return(-1);
}

/*
 * make strings containing i_cookie + r_cookie + msgid
 */
u_char *
isakmp_pindex(index, msgid)
	isakmp_index *index;
	msgid_t *msgid;
{
	static char buf[64];
	u_char *p;
	int i, j;

	memset(buf, 0, sizeof(buf));

	/* copy index */
	p = (u_char *)index;
	for (j = 0, i = 0; i < sizeof(isakmp_index); i++) {
		sprintf((char *)&buf[j], "%02x", p[i]);
		j += 2;
		switch (i) {
		case 7: case 15:
			buf[j++] = ':';
		}
	}

	if (msgid == 0)
		return(buf);

	/* copy msgid */
	p = (u_char *)msgid;
	for (i = 0; i < sizeof(msgid_t); i++) {
		sprintf((char *)&buf[j], "%02x", p[i]);
		j += 2;
	}

	return(buf);
}

/* %%%
 */
void
isakmp_close()
{
	int num = n_soI;

	while (num--)
		(void)close(sockIs[num]);

	return;
}

int
isakmp_send(iph1, buf)
	struct isakmp_ph1 *iph1;
	vchar_t *buf;
{
	struct sockaddr *sa;
	int len;

	sa = iph1->remote;

	/* send HDR;SA to responder */
	if ((len = sendto(sockIs[0], buf->v, buf->l, 0, sa, sizeof(*sa))) < 0) {
		plog("isakmp_send", "sendto (%s)\n", strerror(errno));
		return(-1);
	}

	YIPSDEBUG(DEBUG_NET,
		INET_NTOP(sa, _addr1_);
		plog("isakmp_send",
			"%d bytes message has been sent to %s[%d].\n",
			len, _addr1_, ntohs(_INPORTBYAF(sa))));
	YIPSDEBUG(DEBUG_DNET, pvdump(buf));

	return(0);
}

int
isakmp_resend_ph1(sc)
	struct sched *sc;
{
	struct isakmp_ph1 *iph1 = (struct isakmp_ph1 *)sc->ptr1;
	vchar_t *buf = (vchar_t *)sc->ptr2;

	YIPSDEBUG(DEBUG_STAMP, plog("isakmp_resend_ph1", "resend packet.\n"));
	YIPSDEBUG(DEBUG_SCHED2, plog("isakmp_resend_ph1", "tick over #%s\n",
	    sched_pindex(&sc->index)));

	if (isakmp_send(iph1, buf) < 0)
		return(-1);

	sc->tick = isakmp_timer;

	return(0);
}

/* called as schedule of negotiating isakmp-sa is time over. */
int
isakmp_timeout_ph1(sc)
	struct sched *sc;
{
	struct isakmp_ph1 *iph1 = (struct isakmp_ph1 *)sc->ptr1;
	vchar_t *buf = (vchar_t *)sc->ptr2;

	YIPSDEBUG(DEBUG_STAMP, plog("isakmp_timeout_ph1", "timeout to send.\n"));
	YIPSDEBUG(DEBUG_SCHED2, plog("isakmp_timeout_ph1", "try over #%s\n",
	    sched_pindex(&sc->index)));

	if (isakmp_free_ph1(iph1) < 0)
		return(-1);

	vfree(buf);

	sched_kill(sc);

	return(0);
}

int
isakmp_resend_ph2(sc)
	struct sched *sc;
{
	struct isakmp_ph2 *iph2 = (struct isakmp_ph2 *)sc->ptr1;
	vchar_t *buf = (vchar_t *)sc->ptr2;

	YIPSDEBUG(DEBUG_STAMP, plog("isakmp_resend_ph2", "resend packet.\n"));
	YIPSDEBUG(DEBUG_SCHED2, plog("isakmp_resend_ph2", "tick over #%s\n",
	    sched_pindex(&sc->index)));

	if (isakmp_send(iph2->ph1, buf) < 0)
		return(-1);

	sc->tick = isakmp_timer;

	return(0);
}

/* called as schedule of negotiating ipsec-sa is time over. */
int
isakmp_timeout_ph2(sc)
	struct sched *sc;
{
	struct isakmp_ph2 *iph2 = (struct isakmp_ph2 *)sc->ptr1;
	vchar_t *buf = (vchar_t *)sc->ptr2;

	YIPSDEBUG(DEBUG_STAMP, plog("isakmp_timeout_ph2", "timeout to send.\n"));
	YIPSDEBUG(DEBUG_SCHED2, plog("isakmp_timeout_ph2", "try over #%s\n",
	    sched_pindex(&sc->index)));

	if (isakmp_free_ph2(iph2) < 0)
		return(-1);

	vfree(buf);

	sched_kill(sc);

	return(0);
}

/* called as schedule of isakmp-sa is expired. */
int
isakmp_expire(sc)
	struct sched *sc;
{
	struct isakmp_ph1 *iph1 = (struct isakmp_ph1 *)sc->ptr1;

	YIPSDEBUG(DEBUG_SCHED2, plog("isakmp_expire", "try over #%s\n",
	    sched_pindex(&sc->index)));

	plog("isakmp_expire",
	    "ISAKMP-SA is expired. %s\n", isakmp_pindex(&iph1->index, 0));

	if (iph1->ph2tab.len == 0) {
		if (isakmp_do_expire(iph1) < 0) return(-1);
	} else {
		/* set flag */
		iph1->status |= ISAKMP_EXPIRED;
	}

	sched_kill(sc);

	return(0);
}

int
isakmp_do_expire(iph1)
	struct isakmp_ph1 *iph1;
{
	/* if it's initiator, begin re-negosiation */
	if (iph1->dir == INITIATOR) {
		YIPSDEBUG(DEBUG_STAMP,
		    plog("isakmp_do_expire", "begin ISAKMP-SA re-negosiation.\n"));

		if (isakmp_begin_ident(iph1->cfp) == 0) return(-1);
	}

	/* delete old status record */
	if (isakmp_free_ph1(iph1) < 0) return(-1);

	return(0);
}

/* %%%
 * functions about management of the isakmp status table
 */
/*
 * create new isakmp Phase 1 status record to handle isakmp
 */
struct isakmp_ph1 *
isakmp_new_ph1(index)
	isakmp_index *index;
{
	struct isakmp_ph1 *iph1, *c;

	if (isakmp_ph1byindex(index) != 0) {
		plog("isakmp_new_ph1",
		    "already exists. %s\n", isakmp_pindex(index, 0));
		return(0);
	}

	/* create new iph1 */
	if ((iph1 = CALLOC(sizeof(*iph1), struct isakmp_ph1 *)) == 0) {
		plog("isakmp_new_ph1", "calloc (%s)\n", strerror(errno)); 
		return(0);
	}

	memcpy((char *)&iph1->index, (char *)index, sizeof(*index));
	iph1->status = ISAKMP_PH1_N;

	/* add to phase 1 table */
	iph1->next = (struct isakmp_ph1 *)0;
	iph1->prev = ph1tab.tail;

	if (ph1tab.tail == 0)
		ph1tab.head = iph1;
	else
		ph1tab.tail->next = iph1;
	ph1tab.tail = iph1;
	ph1tab.len++;

	return(iph1);
}

/*
 * free from isakmp Phase 1 status table
 */
int
isakmp_free_ph1(iph1)
	struct isakmp_ph1 *iph1;
{
	struct isakmp_ph1 *c;
	int error = -1;

	/* diagnostics */
	if ((c = isakmp_ph1byindex(&iph1->index)) == 0) {
		plog("isakmp_free_ph1",
			"Why is there no status, %s\n", isakmp_pindex(&iph1->index, 0));
		goto end;
	}

	if (c != iph1) {
		plog("isakmp_free_ph1",
		    "why not equalize, %s\n", isakmp_pindex(&iph1->index, 0));
		goto end;
	}

	if (c->ph2tab.len != 0) {
		plog("isakmp_free_ph1",
		    "why ipsec status is alive, %s\n", isakmp_pindex(&iph1->index, 0));
		goto end;
	}

	/* XXX: free more !? */
	/* if (c->dhp) vfree(c->dhp); because this is static */
	if (c->dhpriv) vfree(c->dhpriv);
	if (c->dhpub) vfree(c->dhpub);
	if (c->dhpub_p) vfree(c->dhpub_p);
	if (c->dhgxy) vfree(c->dhgxy);
	if (c->nonce) vfree(c->nonce);
	if (c->nonce_p) vfree(c->nonce_p);
	if (c->skeyid) vfree(c->skeyid);
	if (c->skeyid_d) vfree(c->skeyid_d);
	if (c->skeyid_a) vfree(c->skeyid_a);
	if (c->skeyid_e) vfree(c->skeyid_e);
	if (c->key) vfree(c->key);
	if (c->hash) vfree(c->hash);
	if (c->iv) vfree(c->iv);
	if (c->ive) vfree(c->ive);
	if (c->ivd) vfree(c->ivd);
	if (c->sa) vfree(c->sa);
	/* if (c->id) vfree(c->id); */
	if (c->id_p) vfree(c->id_p);

	if (c->remote) free(c->remote);

	if (c->isa) {
		if (c->isa->spi) vfree(c->isa->spi);
		(void)free(c->isa);
	}

	/* reap from phase 1 table */
	/* middle */
	if (c->prev && c->next) {
		c->prev->next = c->next;
		c->next->prev = c->prev;
	} else
	/* tail */
	if (c->prev && c->next == 0) {
		c->prev->next = (struct isakmp_ph1 *)0;
		ph1tab.tail = c->prev;
	} else
	/* head */
	if (c->prev == 0 && c->next) {
		c->next->prev = (struct isakmp_ph1 *)0;
		ph1tab.head = c->next;
	} else {
	/* iph2->next == 0 && iph2->prev == 0 */
	/* last one */
		ph1tab.head = (struct isakmp_ph1 *)0;
		ph1tab.tail = (struct isakmp_ph1 *)0;
	}

	ph1tab.len--;

	YIPSDEBUG(DEBUG_STAMP,
	    plog("isakmp_free_ph1", "ISAKMP-SA negotiation is free, %s\n",
	        isakmp_pindex(&iph1->index, 0)));

	(void)free(c);

	error = 0;
end:
	return(error);
}

/*
 * search on table of isakmp Phase 1 status by index.
 */
struct isakmp_ph1 *
isakmp_ph1byindex(index)
	isakmp_index *index;
{
	struct isakmp_ph1 *c;

	for (c = ph1tab.head; c; c = c->next) {
		if (memcmp((char *)&c->index, (char *)index, sizeof(*index)) == 0)
			return(c);
	}

	return(0);
}

/*
 * search on table of isakmp Phase 1 status by i_ck in index.
 */
struct isakmp_ph1 *
isakmp_ph1byindex0(index)
	isakmp_index *index;
{
	struct isakmp_ph1 *c;

	for (c = ph1tab.head; c; c = c->next) {
		if (memcmp((char *)&c->index, (char *)index, sizeof(cookie_t)) == 0)
			return(c);
	}

	return(0);
}

/*
 * search isakmp-sa record established on table of isakmp Phase 1
 * by destination address
 */
struct isakmp_ph1 *
isakmp_ph1byaddr(addr)
	struct sockaddr *addr;
{
	struct isakmp_ph1 *c, *tmp = 0;

	for (c = ph1tab.head; c; c = c->next) {
		if (memcmp((char *)c->remote, addr, sizeof(*addr)) == 0
		 && c->dir == INITIATOR) {
		 	if (! ISSET(c->status, ISAKMP_EXPIRED)) return(c);
			tmp = c;
		}
	}

	return(tmp == 0 ? 0 : tmp);
}

/*
 * create new isakmp Phase 2 status record to handle isakmp in Phase2
 */
struct isakmp_ph2 *
isakmp_new_ph2(iph1, msgid)
	struct isakmp_ph1 *iph1;
	msgid_t *msgid;
{
	struct isakmp_ph2 *iph2 = 0, *c;
	struct isakmp_pl_sa *sa;
	int error = -1;

	/* validation */
	if ((c = isakmp_ph2bymsgid(iph1, msgid)) != 0) {
		plog("isakmp_new_ph2",
		    "already exists, %d.\n", isakmp_pindex(&iph1->index, msgid));
		goto end;
	}

	/* create new iph2 */
	if ((iph2 = CALLOC(sizeof(*iph2), struct isakmp_ph2 *)) == 0) {
		plog("isakmp_new_ph2", "calloc (%s)\n", strerror(errno)); 
		goto end;
	}

	iph2->status = ISAKMP_PH2_N;
	memcpy((char *)&iph2->msgid, (char *)msgid, sizeof(msgid_t));
	iph2->ph1 = iph1;

	iph2->next = (struct isakmp_ph2 *)0;
	iph2->prev = iph1->ph2tab.tail;

	/* add to phase 2 table */
	if (iph1->ph2tab.tail == 0)
		iph1->ph2tab.head = iph2;
	else
		iph1->ph2tab.tail->next = iph2;
	iph1->ph2tab.tail = iph2;
	iph1->ph2tab.len++;

	error = 0;

end:
	if (error) {
		if (iph2) (void)isakmp_free_ph2(iph2);
		iph2 = 0;
	}

	return(iph2);
}

/*
 * free from isakmp Phase 2 status table
 */
int
isakmp_free_ph2(iph2)
	struct isakmp_ph2 *iph2;
{
	struct isakmp_ph1 *iph1;

	/* save isakmp status to check expiration. */
	iph1 = iph2->ph1;

	/* XXX: free more !? */
	if (iph2->nonce) vfree(iph2->nonce);
	if (iph2->nonce_p) vfree(iph2->nonce_p);
	if (iph2->hash) vfree(iph2->hash);
	if (iph2->iv) vfree(iph2->iv);
	if (iph2->ive) vfree(iph2->ive);
	if (iph2->ivd) vfree(iph2->ivd);

	if (iph2->pst) pfkey_free_st(iph2->pst); /* XXX */
	if (iph2->sa) (void)(iph2->sa);
	if (iph2->sa_p) (void)(iph2->sa_p);

	/* reap from phase 2 table */
	/* middle */
	if (iph2->prev && iph2->next) {
		iph2->prev->next = iph2->next;
		iph2->next->prev = iph2->prev;
	} else
	/* tail */
	if (iph2->prev && iph2->next == 0) {
		iph2->prev->next = (struct isakmp_ph2 *)0;
		iph1->ph2tab.tail = iph2->prev;
	} else
	/* head */
	if (iph2->prev == 0 && iph2->next) {
		iph2->next->prev = (struct isakmp_ph2 *)0;
		iph1->ph2tab.head = iph2->next;
	} else {
	/* iph2->next == 0 && iph2->prev == 0 */
		iph1->ph2tab.head = (struct isakmp_ph2 *)0;
		iph1->ph2tab.tail = (struct isakmp_ph2 *)0;
	}

	iph1->ph2tab.len--;

	YIPSDEBUG(DEBUG_STAMP,
	    plog("isakmp_free_ph2", "IPsec-SA negotiation is free, %s\n",
	        isakmp_pindex(&iph1->index, &iph2->msgid)));

	(void)free(iph2);

	/* if isakmp-sa has been expired,
	 * free isakmp-sa and begin to re-negotiation.
	 */
	if (iph1->ph2tab.len == 0 && ISSET(iph1->status, ISAKMP_EXPIRED)) {
		if (isakmp_do_expire(iph1) < 0) return(-1);
	}

	return(0);
}

struct isakmp_ph2 *
isakmp_ph2bymsgid(iph1, msgid)
	struct isakmp_ph1 *iph1;
	msgid_t *msgid;
{
	struct isakmp_ph2 *c;

	for (c = iph1->ph2tab.head; c; c = c->next) {
		if (memcmp((char *)&c->msgid, (char *)msgid, sizeof(msgid_t)) == 0)
			return(c);
	}

	return(0);
}

/* %%%
 * Interface between PF_KEYv2 and ISAKMP
 */
/*
 * receive PFKEYv2 from kernel, and begin either IDENT or QUICK mode.
 * if IDENT mode finished, begin QUICK mode.
 */
int
isakmp_pfkey(pst)
	struct pfkey_st *pst;
{
	struct isakmp_conf *cfp;
	struct isakmp_ph1 *iph1 = 0;
	int error = -1;

	YIPSDEBUG(DEBUG_STAMP, plog("isakmp_pfkey", "begin.\n"));

#if 0
	if ((cfp = isakmp_cfbyaddr(pst->src, pst->dst)) == &cftab) {
		if ((cfp = isakmp_cfbypeer(pst->dst)) == &cftab) {
			plog("isakmp_pfkey", "peer address of ISAKMP couldn't be found.\n");
			goto end;
		}
	}
#else
	if ((cfp = isakmp_cfbypeer(pst->dst)) == &cftab) {
		plog("isakmp_pfkey", "peer address of ISAKMP couldn't be found.\n");
		goto end;
	}
#endif

	/* search isakmp status table by address */
	iph1 = isakmp_ph1byaddr(cfp->remote);

	if (iph1 != 0 && iph1->status == ISAKMP_PH2_N) {
#if defined(SINGLE)
		if (iph1->ph2tab.len == 0) {
#endif
			/* found ISAKMP-SA. */
			YIPSDEBUG(DEBUG_STAMP,
			    plog("isakmp_pfkey", "begin QUICK mode.\n"));

			/* begin quick mode */
			if (isakmp_begin_quick(iph1, pst) < 0) goto end;

#if defined(SINGLE)
		} else {

			/* It's on time to process other negotiation. */
			YIPSDEBUG(DEBUG_NOTIFY,
				plog("isakmp_pfkey",
			        "queued, as on time to negotiate other SA.\n"));
			isakmp_new_queue(pst, cfp->remote);
		}
#endif

	} else
	if (iph1 != 0 && ISSET(iph1->status, ISAKMP_EXPIRED)) {
		/* found ISAKMP-SA, but expired. */
		YIPSDEBUG(DEBUG_STAMP,
		    plog("isakmp_pfkey", "queued, as ISAKMP-SA expired.\n"));
		isakmp_new_queue(pst, cfp->remote);

	} else
	if (iph1 != 0 && (iph1->status != ISAKMP_PH2_N)) {
		/* found ISAKMP-SA, but begin negotiation. */
		YIPSDEBUG(DEBUG_STAMP,
		    plog("isakmp_pfkey", "queued, ISAKMP-SA is on negotiation.\n"));
		isakmp_new_queue(pst, cfp->remote);

	} else {
		/* no ISAKMP-SA found. */
		YIPSDEBUG(DEBUG_STAMP,
		    plog("isakmp_pfkey", "queued, as no phase 1 exists.\n"));

		YIPSDEBUG(DEBUG_STAMP,
		    plog("isakmp_pfkey", "begin IDENT mode.\n"));

		isakmp_new_queue(pst, cfp->remote);

		/* begin ident mode */
		if ((iph1 = isakmp_begin_ident(cfp)) == 0) goto end;
	}

	error = 0;

end:
	return(error);
}

int
isakmp_new_queue(pst, remote)
	struct pfkey_st *pst;
	struct sockaddr *remote;
{
	YIPSDEBUG(DEBUG_SCHED,
	    plog("isakmp_new_queue", "new PFKEY request scheduled.\n"));

	/* add to the schedule to resend, and seve back pointer. */
	pst->sc = sched_add(1, isakmp_pfkey_check,
	                    pfkey_timer, isakmp_pfkey_over,
	                    (caddr_t)pst, (caddr_t)remote);

	return(0);
}

int
isakmp_pfkey_check(sc)
	struct sched *sc;
{
	struct pfkey_st *pst = (struct pfkey_st *)sc->ptr1;
	struct sockaddr *remote = (struct sockaddr *)sc->ptr2;
	struct isakmp_ph1 *iph1;

	YIPSDEBUG(DEBUG_SCHED2,
	    plog("isakmp_pfkey_check",
	        "tick over #%s\n", sched_pindex(&sc->index)));

	iph1 = isakmp_ph1byaddr(remote);
	if (iph1 != 0
	 && iph1->status == ISAKMP_PH2_N
	 && iph1->ph2tab.len == 0) {
		/* found isakmp-sa */
		sched_kill(sc);
		if (isakmp_pfkey(pst) < 0) return(-1);
	} else
		sc->tick = 1;

	return(0);
}

/* called as schedule is time over. */
int
isakmp_pfkey_over(sc)
	struct sched *sc;
{
	struct pfkey_st *pst = (struct pfkey_st *)sc->ptr2;
	struct sockaddr *remote = (struct sockaddr *)sc->ptr2;

	YIPSDEBUG(DEBUG_STAMP,
	    plog("isakmp_pfkey_over", "timeout to negosiate SA.\n"));
	YIPSDEBUG(DEBUG_SCHED2,
	    plog("isakmp_pfkey_over", "try over #%s\n",
	        sched_pindex(&sc->index)));

	plog2(remote, "isakmp_pfkey_over",
		"no IPSEC-SA exchanged, because ISAKMP-SA could't be established.\n");

	/* XXX */
	pfkey_free_st(pst);

	sched_kill(sc);

	return(0);
}

struct isakmp_conf *
isakmp_cfbypeer(remote)
	struct sockaddr *remote;
{
	struct isakmp_conf *cfp;
	u_long mask;

	YIPSDEBUG(DEBUG_MISC,
		INET_NTOP(remote, _addr1_);
		plog("isakmp_cfbypeer",
			"search with remote addr=%s.\n", _addr1_));

	for (cfp = cftab.next; cfp; cfp = cfp->next) {
		if (saddrcmp(remote, cfp->remote) == 0) {
			YIPSDEBUG(DEBUG_MISC,
				INET_NTOP(cfp->remote, _addr1_);
				plog("isakmp_cfbypeer",
					"apply configuration addr=%s[%u]\n",
					_addr1_,
					ntohs(_INPORTBYAF(cfp->remote))));
			return(cfp);
		}
	}

	YIPSDEBUG(DEBUG_MISC,
	    plog("isakmp_cfbypeer", "apply default configuration.\n"));

	return(&cftab); /* default */
}

/*
 * isakmp_cfbyaddr
 *   IN:
 *     src, dst: pointer to the struct sockaddr.
 *   OUT:
 *     pointer to the cftable.
 *     If value is the &cftab, that is default.
 */
struct isakmp_conf *
isakmp_cfbyaddr(src, dst)
	struct sockaddr *src, *dst;
{
	struct isakmp_proxy *prxp;
	u_int masks, maskd;

	YIPSDEBUG(DEBUG_MISC,
		INET_NTOP(src, _addr1_);
		INET_NTOP(dst, _addr2_);
		plog("isakmp_cfbyaddr", "search with address src=%s, dst=%s.\n",
			_addr1_, _addr2_));

	if ((prxp = isakmp_proxybyaddr(src, dst)) == 0) {

		YIPSDEBUG(DEBUG_MISC,
		    plog("isakmp_cfbyaddr", "apply default configuration.\n"));
		return(&cftab); /* default */

	} else {

		YIPSDEBUG(DEBUG_MISC,
			INET_NTOP(prxp->cfp->remote, _addr1_);
			plog("isakmp_cfbyaddr",
				"apply configuration addr=%s[%u]\n",
				_addr1_,
				ntohs(_INPORTBYAF(prxp->cfp->remote))));
		return(prxp->cfp ? prxp->cfp : &cftab);
	}
}

struct isakmp_proxy *
isakmp_proxybyaddr(src, dst)
	struct sockaddr *src, *dst;
{
	struct isakmp_proxy *prxp;
	u_int masks, maskd;

	YIPSDEBUG(DEBUG_MISC,
		INET_NTOP(src, _addr1_);
		INET_NTOP(dst, _addr2_);
		plog("isakmp_proxybyaddr",
			"search with address src=%s, dst=%s.\n",
			_addr1_, _addr2_));

	for (prxp = proxytab; prxp; prxp = prxp->next) {
		masks = (0xffffffff >> (32 - prxp->masks));
		maskd = (0xffffffff >> (32 - prxp->maskd));
		if ((prxp->src & masks) == (*_SADDRBYAF(src) & masks)
		 && (prxp->dst & maskd) == (*_SADDRBYAF(dst) & maskd)) {
			return(prxp);
		}
	}

	return(0);
}

vchar_t *
isakmp_dump_sa()
{
	struct isakmp_ph1 *dst, *src;
	struct admin_com *com;
	vchar_t *buf;
	int tlen;

	tlen = sizeof(struct isakmp_ph1) * ph1tab.len
	     + sizeof(struct admin_com);

	if ((buf = vmalloc(tlen)) == 0) {
		plog("isakmp_dump_sa", "vmalloc(%s)\n", strerror(errno));
		return(0);
	}
	com = (struct admin_com *)buf->v;

	com->cmd = ADMIN_SHOW_SA;
	com->len = tlen;

	dst = (struct isakmp_ph1 *)(buf->v + sizeof(*com));
	for (src = ph1tab.head; src; src = src->next) {
		memcpy((caddr_t)dst, (caddr_t)src, sizeof(*dst));
		dst++;
	}

	return(buf);
}

int
isakmp_pfkey_update(iph2)
	struct isakmp_ph2 *iph2;
{
#if 0
	struct secas *sa = 0;
	static u_int16 said = MAX_USERS_SAID + 1;
	int isakmp_extension = 0;
	u_int32 src, dst;
	int error = -1;

	if (iph2->proxy = EXCHANGE_PROXY) {
		src = iph2->px->src;
		dst = iph2->px->dst;

		sa->masks = iph2->px->masks;
		sa->maskd = iph2->px->masks;

		if ((sa->pxd
		    = CALLOC(sizeof(struct sockaddr), struct sockaddr *)) == 0) {
			plog("isakmp_pfkey_update", "calloc (%s)\n", strerror(errno));
			plog("isakmp_pfkey_update", "calloc (%s)\n", strerror(errno));
			goto end;
		} 
		((struct sockaddr_in *)sa->pxd)->sin_len = sizeof(struct sockaddr_in);
		((struct sockaddr_in *)sa->pxd)->sin_family = PF_INET;
		((struct sockaddr_in *)sa->pxd)->sin_addr.s_addr
		    = htonl(iph2->px->remote);

	} else {
		src = iph2->pst->src.sin_addr.s_addr;
		dst = iph2->pst->dst.sin_addr.s_addr;
	}

	if ((sa = CALLOC(sizeof(struct secas), struct secas *)) == 0) {
		plog("isakmp_pfkey_update", "calloc (%s)\n", strerror(errno));
		goto end;
	} 
	sa->type = iph2->sa->proto_id;
	sa->mode = iph2->sa->enc_t;
	sa->dir = SA_INBOUND;
	memcpy((caddr_t)&sa->spi, iph2->sa->spi->v, sizeof(sa->spi));
	sa->said = said ++;
	sa->extension = isakmp_extension;
	sa->nextpr = 0;

	if ((sa->src
	    = CALLOC(sizeof(struct sockaddr), struct sockaddr *)) == 0) {
		plog("isakmp_pfkey_update", "calloc (%s)\n", strerror(errno));
		goto end;
	} 
	((struct sockaddr_in *)sa->src)->sin_len = sizeof(struct sockaddr_in);
	((struct sockaddr_in *)sa->src)->sin_family = PF_INET;
	((struct sockaddr_in *)sa->src)->sin_addr.s_addr = htonl(src);

	if ((sa->dst
	    = CALLOC(sizeof(struct sockaddr), struct sockaddr *)) == 0) {
		plog("isakmp_pfkey_update", "calloc (%s)\n", strerror(errno));
		goto end;
	} 
	((struct sockaddr_in *)sa->src)->sin_len = sizeof(struct sockaddr_in);
	((struct sockaddr_in *)sa->src)->sin_family = PF_INET;
	((struct sockaddr_in *)sa->src)->sin_addr.s_addr = htonl(dst);

	if ((sa->alg_auth = iph2->sa->hash_t)) {
		if ((sa->key_auth
		    = CALLOC(sizeof(struct secaskey), struct secaskey *)) == 0) {
			plog("isakmp_pfkey_update", "calloc (%s)\n", strerror(errno));
			goto end;
		} 
		sa->key_auth->h.keylen = 8; /* XXX */
		memcpy((caddr_t)&sa->key_auth->key, iph2->sa->keymat->v, 8); /* XXX */
	}

	if ((sa->alg_cipher = iph2->sa->enc_t)) {
		if ((sa->key_cipher
		    = CALLOC(sizeof(struct secaskey), struct secaskey *)) == 0) {
			plog("isakmp_pfkey_update", "calloc (%s)\n", strerror(errno));
			goto end;
		} 
		sa->key_auth->h.keylen = 8; /* XXX */
		memcpy((caddr_t)&sa->key_auth->key, iph2->sa->keymat->v, 8); /* XXX */
	}

	/* pfkey_update */
	if (pfkey_update(sa) < 0)
		goto end;

	error = 0;
end:
	if (sa) free(sa);
	return(error);
#else
	return(0);
#endif
}
